
import sys, os, json, re

from collections import OrderedDict

from microapp import App
from fortlab.kgutils import (UserException, ProgramException, logger, KGName, run_shcmd,
                             KgenConfigParser, traverse)
from fortlab.resolver import kgparse
#from kgconfig import Config
from fortlab.resolver import statements

SRCROOT = os.path.dirname(os.path.realpath(__file__))
KGEN_MACHINE = '%s/../machines'%SRCROOT

class MpiModule(object):

    def __init__(self, header):
        self.header = header

    def resolve(self, request, config):
        pass

def get_MPI_PARAM(node, bag, depth):
    from fortlab.resolver.Fortran2003 import Specification_Part, Type_Declaration_Stmt, Entity_Decl, Parameter_Stmt, Named_Constant_Def, \
        NoMatchError, Module_Stmt, Program_Stmt, Named_Constant_Def_List

    if isinstance(node, Type_Declaration_Stmt):
        if isinstance(node.items[2], Entity_Decl) and node.items[2].items[0].string.upper()==bag['key']:
            pass
    elif isinstance(node, Parameter_Stmt):
        if isinstance(node.items[1], Named_Constant_Def) and node.items[1].items[0].string.upper()==bag['key']:
            bag[bag['key']].append(str(node.items[1].items[1]).replace(' ', ''))
        elif isinstance(node.items[1], Named_Constant_Def_List):
            for item in node.items[1].items:
                if isinstance(item, Named_Constant_Def) and item.items[0].string.upper()==bag['key']:
                    bag[bag['key']].append(str(item.items[1]).replace(' ', ''))


class FortranNameResolver(App):

    _name_ = "resolve"
    _version_ = "0.1.0"

    def __init__(self, mgr):

        self.add_argument("callsitefile", metavar="path", help="callsite file path")
        self.add_argument("--import-source", metavar="srcpath", action="append",
                          help="load source file")
        self.add_argument("--compile-info", metavar="path", help="compiler flags")
        self.add_argument("--keep", metavar="analysis", help="keep analysis")

        ###############################################################
        # Add common options
        ###############################################################
        self.add_argument("-i", "--include-ini", dest="include_ini", action='store', type=str, help="information used for analysis")
        self.add_argument("-e", "--exclude-ini", dest="exclude_ini", action='store', type=str, help="information excluded for analysis")
        self.add_argument("-I", dest="include", action='append', type=str, help="include path information used for analysis")
        self.add_argument("-D", dest="macro", action='append', type=str, help="macro information used for analysis")
        self.add_argument("--outdir", dest="outdir", action='store', type=str, help="path to create outputs")
        self.add_argument("--source", dest="source", action='append', type=str, help="Setting source file related properties")
        self.add_argument("--intrinsic", dest="intrinsic", action='append', type=str, help="Specifying resolution for intrinsic procedures during searching")
        self.add_argument("--machinefile", dest="machinefile", action='store', type=str, help="Specifying machinefile")
        self.add_argument("--debug", dest="debug", action='append', type=str)
        self.add_argument("--logging", dest="logging", action='append', type=str)

        ###############################################################
        # Add extraction options
        ###############################################################

        self.add_argument("--invocation", dest="invocation", action='append', type=str, help="(process, thread, invocation) pairs of kernel for data collection")
        self.add_argument("--data", dest="data", action='append', type=str, help="Control state data generation")
        self.add_argument("--openmp", dest="openmp", action='append', type=str, help="Specifying OpenMP options")
        self.add_argument("--mpi", dest="mpi", action='append', type=str, help="MPI information for data collection")
        self.add_argument("--timing", dest="timing", action='store', type=str, help="Timing measurement information")
        self.add_argument("--prerun", dest="prerun", action='append', type=str, help="prerun commands")
        self.add_argument("--rebuild", dest="rebuild", action='append', type=str, help="rebuild controls")
        self.add_argument("--state-switch", dest="state_switch", action='append', type=str, help="Specifying how to switch orignal sources with instrumented ones.")
        self.add_argument("--kernel-option", dest="kernel_option", action='append', type=str, help="Specifying kernel compiler and linker options")
        self.add_argument("--check", dest="check", action='append', type=str, help="Kernel correctness check information")
        self.add_argument("--verbose", dest="verbose_level", action='store', type='int', help="Set the verbose level for verification output")
        self.add_argument("--add-mpi-frame", dest="add_mpi_frame", action='store', type=str, help="Add MPI frame codes in kernel_driver")
        self.add_argument("--add-cache-pollution", dest="add_cache_pollution", action='store', type=str, help="Add cache pollution frame codes in kernel_driver")

        self.add_argument("--repr-etime", dest="repr_etime", action='append', type=str, help="Specifying elapsedtime representativeness feature flags")
        self.add_argument("--repr-papi", dest="repr_papi", action='append', type=str, help="Specifying papi counter representativeness feature flags")
        self.add_argument("--repr-code", dest="repr_code", action='append', type=str, help="Specifying code coverage representativeness feature flags")

        #self.register_shared("config", help="config object")
        #self.register_shared("trees", help="ast object")
        self.register_forward("analysis", help="analysis object")

        # database
        self.config = {}

        # KGEN version
        self.config['kgen'] = OrderedDict()
        self.config['kgen']['version'] = [ 0, 9, '0' ]

        # machine parameters
        self.config['machine'] = OrderedDict()
        self.config['machine']['inc'] = None
        self.config['machine']['general'] = OrderedDict()
        self.config['machine']['general']['name'] = 'Generic Machine'
        self.config['machine']['general']['id'] = 'generic'
        self.config['machine']['compiler'] = OrderedDict()
        self.config['machine']['compiler']['gnu'] = ''
        self.config['machine']['compiler']['intel'] = ''
        self.config['machine']['compiler']['pgi'] = ''
        self.config['machine']['compiler']['cray'] = ''
        self.config['machine']['variable'] = OrderedDict()
        self.config['machine']['variable']['work_directory'] = os.path.expandvars('${HOME}/kgen_workspace')

        # logging parameters
        self.config['logging'] = OrderedDict()
        self.config['logging']['select'] = OrderedDict()

        # path parameters
        self.config['cwd'] = os.getcwd()
        self.config['path'] = OrderedDict()
        self.config['path']['outdir'] = '.'
        self.config['path']['state'] = 'state'
        self.config['path']['kernel'] = 'kernel'
        self.config['path']['model'] = 'model'
        self.config['path']['coverage'] = 'coverage'
        self.config['path']['etime'] = 'elapsedtime'
        self.config['path']['papi'] = 'papi'

        # source file parameters
        self.config["source"] = OrderedDict()
        self.config['source'] = OrderedDict()
        self.config['source']['isfree'] = None
        self.config['source']['isstrict'] = None
        self.config['source']['alias'] = OrderedDict()
        self.config['source']['file'] = OrderedDict()
        self.config['source']['state'] = []

        # include parameters
        self.config['include'] = OrderedDict()
        self.config['include']['macro'] = OrderedDict()
        self.config['include']['path'] = []
        self.config['include']['type'] = OrderedDict()
        self.config['include']['compiler'] = OrderedDict()
        self.config['include']['import'] = OrderedDict()
        self.config['include']['file'] = OrderedDict()
        self.config['include']['opt'] = None

        # mpi parameters
        self.config['mpi'] = OrderedDict()
        self.config['mpi']['enabled'] = False
        self.config['mpi']['comm'] = None
        self.config['mpi']['logical'] = None
        self.config['mpi']['status_size'] = None
        self.config['mpi']['source'] = None
        self.config['mpi']['any_source'] = None
        self.config['mpi']['header'] = 'mpif.h'
        self.config['mpi']['use_stmts'] = []

        # external tool parameters
        self.config['bin'] = OrderedDict()
        self.config['bin']['pp'] = 'cpp'
        self.config['bin']['cpp_flags'] = '-w -traditional -P'

        self.config['modules'] = OrderedDict()
        self.config['srcfiles'] = OrderedDict()

        self.config['kernel'] = OrderedDict()
        self.config['kernel']['name'] = None

        self.config['mixed'] = OrderedDict()
        self.config['mixed']["varnames"] = OrderedDict()

        self.config['callsite'] = OrderedDict()
        self.config['callsite']['stmts'] = []
        self.config['callsite']['filepath'] = ''
        self.config['callsite']['span'] = (-1, -1)
        self.config['callsite']['namepath'] = ''
#        self.config['callsite']['lineafter'] = -1
        self.config['parentblock'] = OrderedDict()
        self.config['parentblock']['stmt'] = None
        self.config['topblock'] = OrderedDict()
        self.config['topblock']['stmt'] = None
        self.config['topblock']['filepath'] = ''
        self.config['used_srcfiles'] = OrderedDict()
        self.config['kernel_driver'] = OrderedDict()
        self.config['kernel_driver']['name'] = 'kernel_driver'
        self.config['kernel_driver']['callsite_args'] = ['kgen_unit', 'kgen_measure', 'kgen_isverified', 'kgen_filepath']

        # search parameters
        self.config['search'] = OrderedDict()
        self.config['search']['skip_intrinsic'] = True
        self.config['search']['except'] = []
        self.config['search']['promote_exception'] = False

        # resolver parameters
        self.config['custom_resolvers'] = {}

        # exclude parameters
        self.config['exclude'] = OrderedDict()

        # make kernel parameters
        self.config['kernel_option'] = OrderedDict()
        self.config['kernel_option']['FC'] =  ''
        self.config['kernel_option']['FC_FLAGS'] = ''
        self.config['kernel_option']['compiler'] = OrderedDict()
        self.config['kernel_option']['compiler']['add'] = []
        self.config['kernel_option']['compiler']['remove'] = []
        self.config['kernel_option']['linker'] = OrderedDict()
        self.config['kernel_option']['linker']['add'] = []

        # make prerun parameters
        self.config['prerun'] = OrderedDict()
        self.config['prerun']['kernel_build'] = ''
        self.config['prerun']['kernel_run'] = ''
        self.config['prerun']['build'] = ''
        self.config['prerun']['run'] = ''

        # make rebuild parameters
        self.config['rebuild'] = OrderedDict()

        # make cmd parameters
        self.config['cmd_clean'] = OrderedDict()
        self.config['cmd_clean']['cmds'] = ''
        self.config['cmd_build'] = OrderedDict()
        self.config['cmd_build']['cmds'] = ''
        self.config['cmd_run'] = OrderedDict()
        self.config['cmd_run']['cmds'] = ''
        self.config['state_switch'] = OrderedDict()
        self.config['state_switch']['type'] = 'replace'
        self.config['state_switch']['directory'] = ''
        self.config['state_switch']['clean'] = ''

        # exclude parameters
        self.config['exclude'] = OrderedDict()

        
        # openmp parameters
        self.config['openmp'] = OrderedDict()
        self.config['openmp']['enabled'] = False
        self.config['openmp']['critical'] = True
        self.config['openmp']['maxnum_threads'] = 102

        # Fortran parameters
        self.config['fort'] = OrderedDict()
        self.config['fort']['maxlinelen'] = 132

        # program units
        self.config['program_units'] = OrderedDict()

        # debugging parameters
        self.config['debug'] = OrderedDict()
        self.config['debug']['printvar'] = []

        # invocation parameters
        self.config['invocation'] = OrderedDict()
        self.config['invocation']['triples'] = []

        # data parameters
        self.config['data'] = OrderedDict()
        self.config['data']['condition'] = []
        self.config['data']['maxnuminvokes'] = None

        # mpi parameters
        self.config['mpi'] = OrderedDict()
        self.config['mpi']['enabled'] = False
        self.config['mpi']['comm'] = None
        self.config['mpi']['logical'] = None
        self.config['mpi']['status_size'] = None
        self.config['mpi']['source'] = None
        self.config['mpi']['any_source'] = None
        self.config['mpi']['header'] = 'mpif.h'
        self.config['mpi']['use_stmts'] = []

        # add mpi frame code in kernel driver
        self.config['add_mpi_frame'] = OrderedDict()
        self.config['add_mpi_frame']['enabled'] = False
        self.config['add_mpi_frame']['np'] = '2'
        self.config['add_mpi_frame']['mpifc'] = 'mpif90'
        self.config['add_mpi_frame']['mpirun'] = 'mpirun'

        # add mpi frame code in kernel driver
        self.config['add_cache_pollution'] = OrderedDict()
        self.config['add_cache_pollution']['enabled'] = False
        self.config['add_cache_pollution']['size'] = 0

        # plugin parameters
        #self.config['plugin'] = OrderedDict()
        #self.config['plugin']['priority'] = OrderedDict()

        self.config['plugindb'] = OrderedDict()

        # verification parameters
        self.config['verify'] = OrderedDict()
        self.config['verify']['tolerance'] = '1.D-14'
        self.config['verify']['minval'] = '1.D-14'
        self.config['verify']['verboselevel'] = '1'


        # kernel correctness check parameters
        self.config['check'] = OrderedDict()
        #self.config['check']['pert_invar'] = ['*']
        self.config['check']['pert_invar'] = []
        self.config['check']['pert_lim'] = '1.0E-15'

    def read_compile_info(self, cinfo, config):

        for key, value in cinfo.items():
            #if key in [ 'type', 'rename', 'state', 'extern' ]:
            if key in [ 'type', 'macro' ]:
                for option in Inc.options(section):
                    self.config["include"][key][option] = Inc.get(section, option).strip()
            elif key=='import':
                for option in Inc.options(section):
                    self.config["include"][key][option] = Inc.get(section, option).strip()
    #                subflags = OrderedDict()
    #                for subf in Inc.get(section, option).split(','):
    #                    subflags[subf.strip()] = None
    #                self.config["include"][key][option] = subflags
            elif key=='include':
                for option in Inc.options(section):
                    self.config["include"]['path'].append(option.strip())
            elif key=='compiler':
                for option in Inc.options(section):
                    self.config["include"][key][option] = Inc.get(section, option).strip()
            else:
                realpath = os.path.realpath(key)

                if not os.path.exists(realpath):
                    print("WARNING: '%s' does not exist. It may cause failure of KGen analysis." % realpath)

                if realpath not in self.config["include"]['file']:
                    self.config["include"]['file'][realpath] = OrderedDict()
                    self.config["include"]['file'][realpath]['path'] = ['.']
                    self.config["include"]['file'][realpath]['compiler'] = None
                    self.config["include"]['file'][realpath]['compiler_options'] = None
                    self.config["include"]['file'][realpath]['macro'] = OrderedDict()
                    self.config["include"]['file'][realpath]['srcbackup'] = []

                for infotype, infovalue in value.items():
                    if infotype=='include':
                        self.config["include"]['file'][realpath]['path'].extend(infovalue)
                    elif infotype in [ 'compiler', 'options', "openmp", "srcbackup" ]:
                        self.config["include"]['file'][realpath][infotype] = infovalue
                    else:
                        for mkey, mvalue in infovalue:
                            self.config["include"]['file'][realpath]['macro'][mkey] = mvalue

        # dupulicate paths per each alias
        newpath = set()
        for path in self.config['include']['path']:
            newpath.add(path)
            for p1, p2 in self.config['source']['alias'].items():
                if path.startswith(p1):
                    newpath.add(p2+path[len(p1):])
                elif path.startswith(p2):
                    newpath.add(p1+path[len(p2):])
        self.config['include']['path'] = list(newpath)

        newfile =  OrderedDict()
        for path, value in self.config['include']['file'].items():
            newfile[path] = value
            for p1, p2 in self.config['source']['alias'].items():
                if path.startswith(p1):
                    newpath = p2+path[len(p1):]
                    newfile[newpath] = copy.deepcopy(value)
                elif path.startswith(p2):
                    newpath = p1+path[len(p2):]
                    newfile[newpath] = copy.deepcopy(value)
        self.config['include']['file'] = newfile

        for path, value in self.config['include']['file'].items():
            if "path" in value:
                newpath = set()
                for path in value['path']:
                    newpath.add(path)
                    for p1, p2 in self.config['source']['alias'].items():
                        if path.startswith(p1):
                            newpath.add(p2+path[len(p1):])
                        elif path.startswith(p2):
                            newpath.add(p1+path[len(p2):])
                value['path'] = list(newpath)


    def perform(self, args):

        from fortlab.resolver.kgsearch import f2003_search_unknowns
        import fortlab.resolver.kganalyze as kganalyze

        if args.compile_info:
            cinfo = args.compile_info["_"]

            if isinstance(cinfo, str):
                # read json file
                with open(cinfo) as f:
                    self.read_compile_info(json.load(f), self.config)

            elif isinstance(cinfo, dict):
                self.read_compile_info(cinfo, self.config)

            else:
                print("Wrong compile-info type: %s" % type(cinfo))
                sys.exit(-1)

        # process default flags
        self._process_default_flags(args)

        # process extraction flags
        self._process_extract_flags(args)

#        # process representativeness flags
#        self._process_repr_flags(args)

        if self.config['mpi']['enabled']:
            self.collect_mpi_params()

        # preprocess if required
        if args.import_source:
            for iarg in args.import_source:
                iarg = iarg["_"]
                kgparse.SrcFile(iarg["_"], self.config)

        callsite = args.callsitefile["_"].split(':', 1)
        if not os.path.isfile(callsite[0]):
            raise UserException('ERROR: callsite file, "%s" can not be found.' % callsite[0])

        # set callsite filepath
        self.config["callsite"]['filepath'] = os.path.realpath(callsite[0])

        # set namepath if exists in command line argument
        if len(callsite)==2:
            self.config["callsite"]['namepath'] = callsite[1].lower()

        elif len(callsite)>2:
            raise UserException('ERROR: Unrecognized call-site information(Syntax -> filepath[:subprogramname]): %s'%str(callsite))

        # read source file that contains callsite stmt
        cs_file = kgparse.SrcFile(self.config["callsite"]["filepath"], self.config)
        if len(self.config["callsite"]['stmts'])==0:
            raise UserException('Can not find callsite')

        # add geninfo to ancestors
        ancs = self.config["callsite"]['stmts'][0].ancestors()

        self.add_geninfo_ancestors(self.config["callsite"]['stmts'][0])

        # populate parent block parameters
        self.config["parentblock"]['stmt'] = ancs[-1]

        # populate top block parameters
        self.config["topblock"]['stmt'] = ancs[0]
        self.config["topblock"]['filepath'] = os.path.realpath(self.config["topblock"]['stmt'].reader.id)

        # resolve
        for cs_stmt in self.config["callsite"]['stmts']:
            #resolve cs_stmt
            f2003_search_unknowns(cs_stmt, cs_stmt.f2003, self.config)
            if hasattr(cs_stmt, 'unknowns'):
                for uname, req in cs_stmt.unknowns.items():
                    cs_stmt.resolve(req, self.config)
                    if not req.res_stmts:
                        raise ProgramException('Resolution fail.')
            else:
                logger.warn('Stmt does not have "unknowns" attribute: %s'%str(cs_stmt)) 

        # update state info of callsite and its upper blocks
        kganalyze.update_state_info(self.config["parentblock"]['stmt'], self.config)

        # update state info of modules
        for modname, moddict in self.config["modules"].items():
            modstmt = moddict['stmt']
            if modstmt != self.config["topblock"]['stmt']:
                kganalyze.update_state_info(moddict['stmt'], self.config)

#        def jsondefault(o):
#            with open("test.json", "w") as f:
#                ll = []
#                import pdb; pdb.set_trace()
#                for i, c in enumerate(o.content):
#                    print("TTTT", i)
#                    ll.append(json.dump(f, c))

        if args.keep:
            print("Saving AST feature is not supported yet.")
#            output_file = open(args.save["_"], 'w')
#            json.dump(self.config, output_file, indent=4, separators=(',', ': '), default=jsondefault)
#            output_file.write('\n')
#            output_file.close()

            #opts = ["@analysis", "-o", args.["_"]]
            #ret, fwds = self.run_subapp("dict2json", opts, forward={"analysis": self.config})
            #assert ret == 0, "dict2json returned non-zero code during analysis saving."

        self.add_forward(analysis=self.config)

#
#    def add_geninfo_ancestors(self, stmt):
#        from block_statements import EndStatement
#
#        ancs = stmt.ancestors()
#
#        prevstmt = stmt
#        prevname = None
#
#        for anc in reversed(ancs):
#            if not hasattr(anc, 'geninfo'):
#                anc.geninfo = OrderedDict()
#            if len(anc.content)>0 and isinstance(anc.content[-1], EndStatement) and \
#                not hasattr(anc.content[-1], 'geninfo'):
#                anc.content[-1].geninfo = OrderedDict()
#
#            if prevname:
#                dummy_req = kgparse.ResState(kgparse.KGGenType.STATE_IN, kgutils.KGName(prevname), None, [anc])
#                dummy_req.res_stmts = [ prevstmt ]
#                anc.check_spec_stmts(dummy_req.uname, dummy_req)
#
#            if hasattr(anc, 'name'): prevname = anc.name
#            else: prevname = None
#            prevstmt = anc
#
    def add_geninfo_ancestors(self, stmt):
        from fortlab.resolver.block_statements import EndStatement

        ancs = stmt.ancestors()

        prevstmt = stmt
        prevname = None

        for anc in reversed(ancs):
            if not hasattr(anc, 'geninfo'):
                anc.geninfo = OrderedDict()
            if len(anc.content)>0 and isinstance(anc.content[-1], EndStatement) and \
                not hasattr(anc.content[-1], 'geninfo'):
                anc.content[-1].geninfo = OrderedDict()

            if prevname:
                dummy_req = kgparse.ResState(kgparse.KGGenType.STATE_IN, KGName(prevname), None, [anc])
                dummy_req.res_stmts = [ prevstmt ]
                anc.check_spec_stmts(dummy_req.uname, dummy_req, self.config)

            if hasattr(anc, 'name'): prevname = anc.name
            else: prevname = None
            prevstmt = anc

    def collect_mpi_params(self):
        #from parser.api import parse, walk
        from fortlab.resolver import api

        if self.config['mpi']['enabled']:
            # get path of mpif.h
            mpifpath = ''
            if os.path.isabs(self.config['mpi']['header']):
                if os.path.exists(self.config['mpi']['header']):
                    mpifpath = self.config['mpi']['header']
                else:
                    raise UserException('Can not find %s'%self.config['mpi']['header'])
            else:
                for p in self.config['include']['path']:
                    fp = os.path.join(p, self.config['mpi']['header'])
                    if os.path.exists(fp):
                        mpifpath = fp
                        break
                if not mpifpath:
                    for incpath, incdict in self.config['include']['file'].items():
                        for p in incdict['path']:
                            fp = os.path.join(p, self.config['mpi']['header'])
                            if os.path.exists(fp):
                                mpifpath = fp
                                break
                        if mpifpath: break

            # collect required information
            if mpifpath:
                try:
                    with open(mpifpath, 'r') as f:
                        filelines = f.read().split('\n')
                        lines = '\n'.join(handle_include(os.path.dirname(mpifpath), filelines, self.config))
                        #reader = FortranStringReader(lines)
                    tree = api.parse(lines, ignore_comments=True, analyze=False, isfree=True, isstrict=False, include_dirs=None, source_only=None )
                    for stmt, depth in api.walk(tree, -1):
                        stmt.parse_f2003()

                    #import pdb; pdb.set_trace()
                    #spec = Specification_Part(reader)
                    bag = {}
                    config_name_mapping = [
                        ('comm', 'MPI_COMM_WORLD'),
                        ('logical', 'MPI_LOGICAL'),
                        ('status_size', 'MPI_STATUS_SIZE'),
                        ('any_source', 'MPI_ANY_SOURCE'),
                        ('source', 'MPI_SOURCE'),
                        ]
                    for config_key, name in config_name_mapping:
                        if config_key not in self.config['mpi'] or self.config['mpi'][config_key] is None:
                            for stmt, depth in api.walk(tree, -1):
                                bag['key'] = name
                                bag[name] = []
                                if hasattr(stmt, 'f2003'):
                                    traverse(stmt.f2003, get_MPI_PARAM, bag, subnode='content')
                                    if len(bag[name]) > 0:
                                        self.config['mpi'][config_key] = bag[name][-1]
                                        break

                    for config_key, name in config_name_mapping:
                        if config_key not in self.config['mpi'] or self.config['mpi'][config_key] is None:
                            raise UserException('Can not find {name} in mpif.h'.format(name=name))

                except UserException:
                    raise  # Reraise this exception rather than catching it below
                except Exception as e:
                    raise UserException('Error occurred during reading %s.'%mpifpath)
            else:
                raise UserException('Can not find mpif.h. Please provide a path to the file')


    def _process_default_flags(self, opts):

        # check if exists fpp or cpp
        output = ''
        try:
            out, err, retcode = run_shcmd('which cpp', show_error_msg=False)
            output = out.strip()
        except Exception as e: pass
        if output.endswith(b'cpp'):
            self.config["bin"]['pp'] = str(output, "utf-8")
        else:
            print('ERROR: cpp is not found.')
            sys.exit(-1)

        # parsing intrinsic skip option
        if opts.intrinsic:
            subflags = []
            for line in opts.intrinsic:
                line = line["_"]
                subflags.extend(line.split(','))

            for subf in subflags:
                if subf and subf.find('=')>0:
                    key, value = subf.split('=')
                    if key=='except':
                        self.config['search']['except'].extend(value.split(';'))
                    elif key=='add_intrinsic':
                        Intrinsic_Procedures.extend([name.lower() for name in value.split(';')])
                    else:
                        raise UserException('Unknown intrinsic sub option: %s' % subf)
                else:
                    if subf=='skip':
                        self.config['search']['skip_intrinsic'] = True
                    elif subf=='noskip':
                        self.config['search']['skip_intrinsic'] = False
                    else:
                        raise UserException('Unknown intrinsic option(s) in %s' % subf)

        # parsing include parameters
        if opts.include:
            for inc in opts.include:
                inc = inc["_"]
                inc_eq = inc.split('=')
                if len(inc_eq)==1:
                    for inc_colon in inc_eq[0].split(':'):
                        self.config['include']['path'].append(inc_colon)
                if len(inc_eq)==1:
                    for inc_colon in inc_eq[0].split(':'):
                        self.config['include']['path'].append(inc_colon)
                elif len(inc_eq)==2:
                    # TODO: support path for each file
                    pass
                else: raise UserException('Wrong format include: %s'%inc)

        if opts.include_ini:
            self.config['include']['opt'] = opts.include_ini["_"]

        if opts.exclude_ini:
            self.process_exclude_option(opts.exclude_ini["_"], self.config['exclude'])

        # parsing macro parameters
        if opts.macro:
            for line in opts.macro:
                line = line["_"]
                for macro in line.split(','):
                    macro_eq = macro.split('=')
                    if len(macro_eq)==1:
                        self.config['include']['macro'][macro_eq[0]] = None
                    elif len(macro_eq)==2:
                        self.config['include']['macro'][macro_eq[0]] = macro_eq[1]
                    else: raise UserException('Wrong format include: %s'%inc)

        files = None
        if opts.source:

            isfree = None
            isstrict = None

            for line in opts.source:
                line = line["_"]
                flags = OrderedDict()
                for subflag in line.split(','):
                    if subflag.find('=')>0:
                        key, value = subflag.split('=')
                        if key=='format':
                            if value == 'free':
                                isfree = True
                            elif value == 'fixed':
                                isfree = False
                            self.config['source']['isfree'] = isfree
                        elif key=='strict':
                            if value == 'yes':
                                isstrict = True
                            elif value == 'no':
                                isstrict = False
                            self.config['source']['isstrict'] = isstrict

            for line in opts.source["_"]:
                flags = OrderedDict()
                for subflag in line.split(','):
                    if subflag.find('=')>0:
                        key, value = subflag.split('=')
                        if key=='file':
                            flags[key] = value.split(':')
                        elif key=='alias':
                            p1, p2 = value.split(':')
                            if p1.endswith('/'): p1 = p1[:-1]
                            if p2.endswith('/'): p2 = p2[:-1]
                            self.config['source']['alias'][p1] = p2
                        elif key=='state':
                            for path in value.split(':'):
                                if os.path.exists(path):
                                    realpath = os.path.realpath(path)
                                    self.config['source'][key].append(realpath)
                                else:
                                    raise UserException('%s does not exist.'%os.path.realpath(path))
                        else:
                            flags[key] = value
                    else:
                        flags[subflag] = None

                isfree = None
                isstrict = None

            for line in opts.source["_"]:
                flags = OrderedDict()
                for subflag in line.split(','):
                    if subflag.find('=')>0:
                        key, value = subflag.split('=')
                        if key=='file':
                            flags[key] = value.split(':')
                        elif key=='alias':
                            p1, p2 = value.split(':')
                            if p1.endswith('/'): p1 = p1[:-1]
                            if p2.endswith('/'): p2 = p2[:-1]
                            self.config['source']['alias'][p1] = p2
                        elif key=='state':
                            for path in value.split(':'):
                                if os.path.exists(path):
                                    realpath = os.path.realpath(path)
                                    self.config['source'][key].append(realpath)
                                else:
                                    raise UserException('%s does not exist.'%os.path.realpath(path))
                        else:
                            flags[key] = value
                    else:
                        flags[subflag] = None

                isfree = None
                isstrict = None


        # parsing debugging options
        # syntax: a.b.c=d,e,f
        if opts.debug:
            for dbg in opts.debug:
                dbg = dbg["_"]
                param_path, value = dbg.split('=')
                param_split = param_path.lower().split('.')
                value_split = value.lower().split(',')
                curdict = self.config['debug']
                for param in param_split[:-1]:
                    curdict = curdict[param]
                exec('curdict[param_split[-1]] = value_split')

        # parsing logging options
        if opts.logging:
            for log in opts.logging:
                log = log["_"]
                param_path, value = log.split('=')
                param_split = param_path.lower().split('.')
                value_split = value.lower().split(',')
                curdict = self.config['logging']
                for param in param_split[:-1]:
                    curdict = curdict[param]
                exec('curdict[param_split[-1]] = value_split')

        if opts.outdir:
            self.config['path']['outdir'] = opts.outdir["_"]

        if opts.machinefile:
            if os.path.exists(opts.machinefile["_"]):
                self.config['machine']['inc'] = KgenConfigParser(allow_no_value=True)
                inc.read(opts.machinefile["_"])
            else:
                print('WARNING: "%s" machine file does not exist.'%opts.machinefile["_"])
        else:
            self.config['machine']['inc'] = self.find_machine()

        if self.config['machine']['inc']:
            self.read_machinefile(self.config['machine'], self.config['prerun'])

        if not os.path.exists(self.config['machine']['variable']['work_directory']):
            os.makedirs(self.config['machine']['variable']['work_directory'])

        # create state directories and change working directory
        if not os.path.exists(self.config['path']['outdir']):
            os.makedirs(self.config['path']['outdir'])
        os.chdir(self.config['path']['outdir'])

    def _process_extract_flags(self, opts):

        # parsing invocation parameters
        if opts.invocation:
            for line in opts.invocation:
                line = line["_"]
                for invocation in line.split(','):
                    t = invocation.split(':')
                    if len(t) != 3:
                        raise UserException('Wrong invocation syntax: expected <mpi ranks>:<openmp numbers>:invocations but used %s'%invocation)

                    triple = []
                    for pair in t:
                        r = pair.split('-')
                        if len(r)==1:
                            triple.append((r[0],r[0]))
                        elif len(r)==2:
                            triple.append(r)
                        else:
                            raise UserException('Wrong invocation syntax: expected a single number or "number-number" format but used %s'%pair)
                    try:
                        int(triple[2][0])
                        int(triple[2][1])
                    except:
                        raise UserException('The last item in invocation triple should be number.')
                    self.config['invocation']['triples'].append(triple)

        # parsing data parameters
        if opts.data:
            for line in opts.data:
                line = line["_"]
                for data in line.split(','):
                    key, value = data.split('=', 1)
                    if key == 'condition':
                        t = value.strip().split(':')
                        if len(t) == 1:
                            self.config['data']['condition'].append(('and', t[0]))
                        elif len(t) == 2:
                            #if t[0] in ('set', 'or', 'and'):
                            if t[0] in ('and'): # supports "and" only
                                self.config['data']['condition'].append((t[0], t[1]))
                            else:
                                raise UserException('Unknown condition action type: %s' % t[0])
                        else:
                            raise UserException('Wrong number of condition subvalues: %s' % value)
                    elif key == 'maxnuminvokes':
                        self.config['data']['maxnuminvokes'] = int(value)
                    else:
                        raise UserException('Unknown data option: %s' % key)
        # parsing OpenMP parameters
        if opts.openmp:
            self.config['openmp']['enabled'] = True
            for line in opts.openmp:
                line = line["_"]
                for openmp in line.split(','):
                    if openmp=='enable':
                        pass
                    else:
                        key, value = openmp.split('=')
                        if key=='kernel-in-critical-region':
                            if value=='no':
                                self.config['openmp']['critical'] = False
                        elif key=='omp_num_threads':
                            if isinstance(value, str) and value.isdigit():
                                self.config['openmp']['maxnum_threads'] = int(value)
                        else:
                            raise UserException('Unknown OpenMP option: %s' % openmp)

        # parsing MPI parameters
        if opts.mpi:
            self.config['mpi']['enabled'] = True

            if "module" not in self.config["custom_resolvers"]:
                self.config["custom_resolvers"]["module"] = {}
            if "mpi" not in self.config["custom_resolvers"]["module"]:
                self.config["custom_resolvers"]["module"]["mpi"] = False

            for line in opts.mpi:
                line = line["_"]
                for mpi in line.split(','):
                    if mpi=='enable':
                        pass
                    else:
                        key, value = mpi.split('=', 1)
                        if key=='comm':
                            self.config['mpi'][key] = value
                        elif key=='use':
                            mod_name, identifier = value.split(':')
                            self.config['mpi']['use_stmts'].append((mod_name, [identifier]))
                        elif key=='ranks':
                            print('ranks subflag for mpi is not supported. Please use invocation flag instead')
                            sys.exit(-1)
                            #self.config['mpi'][key] = value.split(':')
                            #self.config['mpi']['size'] = len(self.config['mpi'][key])
                        elif key=='header':
                            self.config['mpi'][key] = value
                        elif key=='build_resolver':
                            if value.lower() in ["true", "yes"]:
                                self.config["custom_resolvers"]["module"]["mpi"] = \
                                MpiModule(self.config["mpi"]["header"])
                        else:
                            raise UserException('Unknown MPI option: %s' % mpi)

        # parsing kernel makefile parameters
        if opts.prerun:
            for line in opts.prerun:
                line = line["_"]
                for comp in line.split(','):
                    key, value = comp.split('=', 1)
                    if key in [ 'build', 'run', 'kernel_build', 'kernel_run' ] :
                        self.config['prerun'][key] = dequote(value)
                    else:
                        raise UserException('Unknown prerun option: %s' % comp)

        if opts.rebuild:
            for line in opts.rebuild:
                line = line["_"]
                for comp in line.split(','):
                    self.config['rebuild'][comp] = True
#
#        if opts.cmd_clean:
#            self.config['cmd_clean']['cmds'] = dequote(opts.cmd_clean["_"])
#
#
#        if opts.cmd_build:
#            self.config['cmd_build']['cmds'] = dequote(opts.cmd_build["_"])
#
#        if opts.cmd_run:
#            self.config['cmd_run']['cmds'] = dequote(opts.cmd_run["_"])

        if opts.state_switch:
            for line in opts.state_switch:
                line = line["_"]
                for run in line.split(','):
                    key, value = run.split('=', 1)
                    if key in [ 'directory', 'type', 'clean', 'script' ] :
                        self.config['state_switch'][key] = dequote(value)
                    else:
                        raise UserException('Unknown state-switch option: %s' % run)

        if opts.kernel_option:
            for line in opts.kernel_option:
                line = line["_"]
                for kopt in line.split(','):
                    split_kopt = kopt.split('=', 1)
                    if len(split_kopt)==1:
                        self.config['kernel_option']['compiler']['add'].append(dequote(split_kopt[0]))
                    elif len(split_kopt)==2:
                        if split_kopt[0] in [ 'FC', 'FC_FLAGS' ]:
                            self.config['kernel_option'][split_kopt[0]] = dequote(split_kopt[1])
                        elif split_kopt[0] in ('add', 'remove'):
                            self.config['kernel_option']['compiler'][split_kopt[0]].extend(dequote(split_kopt[1]).split(':'))
                        elif split_kopt[0]=='link':
                            self.config['kernel_option']['linker']['add'].extend(dequote(split_kopt[1]).split(':'))
                        else:
                            raise UserException('Unknown state-switch option: %s' % kopt)

        # kernel correctness checks
        if opts.check:
            for line in opts.check:
                line = line["_"]
                for checkparams in line.split(','):
                    key, value = checkparams.split('=', 1)
                    key = key.lower()
                    value = value.lower()
                    if key=='pert_invar':
                        self.config['check'][key] = value.split(':')
                    elif key=='pert_lim':
                        self.config['check'][key] = value
                    elif key=='tolerance':
                        self.config['verify'][key] = value
                    else:
                        print('WARNING: %s is not supported check parameter'%key)

        # parsing logging options
        if opts.verbose_level:
            self.config['verify']['verboselevel'] = str(opts.verbose_level["_"])

        # mpi frame code in kernel driver
        if opts.add_mpi_frame:
            for checkparams in opts.add_mpi_frame["_"].split(','):
                sparam = checkparams.split('=')
                if len(sparam) == 2:
                    key, value = sparam
                    key = key.lower()
                    if key in ['np', 'mpirun', 'mpifc']:
                        self.config['add_mpi_frame']['enabled'] = True
                        self.config['add_mpi_frame'][key] = value
                    else:
                        print('WARNING: %s is not supported add_mpi_frame parameter'%key)
                elif len(sparam) == 1:
                    if sparam[0] == "enabled":
                        self.config['add_mpi_frame']['enabled'] = True
                    else:
                        print('WARNING: %s is not supported add_mpi_frame parameter'%sparam[0])

        # cache pollution frame code in kernel driver
        if opts.add_cache_pollution:
            if opts.add_cache_pollution["_"].isdigit():
                self.config['add_cache_pollution']['enabled'] = True
                self.config['add_cache_pollution']['size'] = int(opts.add_cache_pollution["_"])
            else:
                print('WARNING: %s is not supported add_cache_pollution parameter'%opts.add_cache_pollution["_"])

#
#    def _process_repr_flags(self, opts):
#
#        if opts.repr_etime:
#            for line in opts.repr_etime:
#                line = line["_"]
#                for eopt in line.split(','):
#                    split_eopt = eopt.split('=', 1)
#                    if len(split_eopt)==1:
#                        if split_eopt[0] == 'enable':
#                            self.config['model']['types']['etime']['enabled'] = True
#                        elif split_eopt[0] == 'disable':
#                            self.config['model']['types']['etime']['enabled'] = False
#                        else:
#                            raise UserException('Unknown elapsed-time flag option: %s' % eopt)
#                    elif len(split_eopt)==2:
#
#                        self.config['model']['types']['etime']['enabled'] = True
#
#                        if split_eopt[0] in [ 'minval', 'maxval' ]:
#                            self.config['model']['types']['etime'][split_eopt[0]] = float(split_eopt[1])
#                        elif split_eopt[0] in ('nbins', 'ndata'):
#                            self.config['model']['types']['etime'][split_eopt[0]] = int(split_eopt[1])
#                        elif split_eopt[0] in ('timer', ):
#                            self.config['model']['types']['etime'][split_eopt[0]] = split_eopt[1]
#                        else:
#                            raise UserException('Unknown elapsed-time flag option: %s' % eopt)
#
#        if opts.repr_papi:
#            for line in opts.repr_papi:
#                line = line["_"]
#                for popt in line.split(','):
#                    split_popt = popt.split('=', 1)
#                    if len(split_popt)==1:
#                        if split_popt[0] == 'enable':
#                            self.config['model']['types']['papi']['enabled'] = True
#                        elif split_popt[0] == 'disable':
#                            self.config['model']['types']['papi']['enabled'] = False
#                        else:
#                            raise UserException('Unknown papi-counter flag option: %s' % popt)
#                    elif len(split_popt)==2:
#
#                        self.config['model']['types']['papi']['enabled'] = True
#
#                        if split_popt[0] in [ 'minval', 'maxval', 'header', 'static', 'dynamic', 'event' ]:
#                            self.config['model']['types']['papi'][split_popt[0]] = split_popt[1]
#                        elif split_popt[0] in ('nbins', 'ndata'):
#                            self.config['model']['types']['papi'][split_popt[0]] = int(split_popt[1])
#                        else:
#                            raise UserException('Unknown papi-counter flag option: %s' % popt)
#        if opts.repr_code:
#            for line in opts.repr_code:
#                line = line["_"]
#                for copt in line.split(','):
#                    split_copt = copt.split('=', 1)
#                    if len(split_copt)==1:
#                        if split_copt[0] == 'enable':
#                            self.config['model']['types']['code']['enabled'] = True
#                        elif split_copt[0] == 'disable':
#                            self.config['model']['types']['code']['enabled'] = False
#                        else:
#                            raise UserException('Unknown code-coverage flag option: %s' % copt)
#                    elif len(split_copt)==2:
#
#                        self.config['model']['types']['code']['enabled'] = True
#
#                        if split_copt[0] in [ 'percentage' ]:
#                            self.config['model']['types']['code'][split_copt[0]] = float(split_copt[1])
#                        elif split_copt[0] in [ 'filter' ]:
#                            self.config['model']['types']['code'][split_copt[0]] = split_copt[1].strip().split(':')
#                        elif split_copt[0] in ( 'ndata' ):
#                            self.config['model']['types']['code'][split_popt[0]] = int(split_copt[1])
#                        else:
#                            raise UserException('Unknown code-coverage flag option: %s' % copt)
#
#            # enable coverage feature at extractor
#            #if self.config['model']['types']['code']['enabled']:
#            #    self.config['plugin']['priority']['ext.coverage'] = '%s/plugins/coverage'%KGEN_EXT

    def get_exclude_actions(self, section_name, *args ):
        if section_name=='namepath':
            if len(args)<1: return []

            if section_name in self.config["exclude"]:
                options = self.config["exclude"][section_name]
                for pattern, actions in options.items():
                    if match_namepath(pattern, args[0]):
                        return actions
            return []
        else:
            UserException('Not supported section name in exclusion input file: %s'%section)

    def process_exclude_option(self, exclude_option, excattrs):

        # collect exclude configuration information
        Exc = KgenConfigParser(allow_no_value=True)
        #Exc.optionxform = str
        Exc.read(exclude_option)
        if len(Exc.sections())>0:
            for section in Exc.sections():
                lsection = section.lower().strip()
                if lsection=='common':
                    print('ERROR: a section of "common" is discarded in INI file for exclusion. Please use "namepath" section instead')
                    sys.exit(-1)

                excattrs[lsection] = OrderedDict()
                for option in Exc.options(section):
                    loption = option.lower().strip()
                    excattrs[lsection][loption] = Exc.get(section, option).strip().split('=')
        else:
            UserException('Can not find exclude file: %s'%exclude_option)



    def process_include_option(self):

        incattrs = self.config["include"]

        if incattrs['opt']:
            self.config["includefile"] = os.path.basename(incattrs['opt'])
            shutil.copy(incattrs['opt'], self.config["path"]['outdir'])

        # collect include configuration information
        Inc = KgenConfigParser(allow_no_value=True)
        #Inc.optionxform = str
        Inc.read('%s/%s'%(self.config["path"]['outdir'], self.config["includefile"]))
        for section in Inc.sections():
            lsection = section.lower().strip()
            #if lsection in [ 'type', 'rename', 'state', 'extern' ]:
            if lsection in [ 'type', 'macro' ]:
                for option in Inc.options(section):
                    incattrs[lsection][option] = Inc.get(section, option).strip()
            elif lsection=='import':
                for option in Inc.options(section):
                    incattrs[lsection][option] = Inc.get(section, option).strip()
    #                subflags = OrderedDict()
    #                for subf in Inc.get(section, option).split(','):
    #                    subflags[subf.strip()] = None
    #                incattrs[lsection][option] = subflags
            elif lsection=='include':
                for option in Inc.options(section):
                    incattrs['path'].append(option.strip())
            elif lsection=='compiler':
                for option in Inc.options(section):
                    incattrs[lsection][option] = Inc.get(section, option).strip()
            elif os.path.isfile(section):
                realpath = os.path.realpath(section)
                if realpath not in incattrs['file']:
                    incattrs['file'][realpath] = OrderedDict()
                    incattrs['file'][realpath]['path'] = ['.']
                    incattrs['file'][realpath]['compiler'] = None
                    incattrs['file'][realpath]['compiler_options'] = None
                    incattrs['file'][realpath]['macro'] = OrderedDict()
                for option in Inc.options(section):
                    if option=='include':
                        pathlist = Inc.get(section, option).split(':')
                        incattrs['file'][realpath]['path'].extend(pathlist)
                    elif option in [ 'compiler', 'compiler_options' ]:
                        incattrs['file'][realpath][option] = Inc.get(section, option)
                    else:
                        incattrs['file'][realpath]['macro'][option] = Inc.get(section, option)
            else:
                pass
                #print '%s is either not suppored keyword or can not be found. Ignored.' % section
        # dupulicate paths per each alias
        newpath = set()
        for path in self.config['include']['path']:
            newpath.add(path)
            for p1, p2 in self.config['source']['alias'].items():
                if path.startswith(p1):
                    newpath.add(p2+path[len(p1):])
                elif path.startswith(p2):
                    newpath.add(p1+path[len(p2):])
        self.config['include']['path'] = list(newpath)

        newfile =  OrderedDict()
        for path, value in self.config['include']['file'].items():
            newfile[path] = value
            for p1, p2 in self.config['source']['alias'].items():
                if path.startswith(p1):
                    newpath = p2+path[len(p1):]
                    newfile[newpath] = copy.deepcopy(value)
                elif path.startswith(p2):
                    newpath = p1+path[len(p2):]
                    newfile[newpath] = copy.deepcopy(value)
        self.config['include']['file'] = newfile

        for path, value in self.config['include']['file'].items():
            if "path" in value:
                newpath = set()
                for path in value['path']:
                    newpath.add(path)
                    for p1, p2 in self.config['source']['alias'].items():
                        if path.startswith(p1):
                            newpath.add(p2+path[len(p1):])
                        elif path.startswith(p2):
                            newpath.add(p1+path[len(p2):])
                value['path'] = list(newpath)


    def find_machine(self):

        import glob

        # find machine files
        inc = None
        for path in glob.glob('%s/*'%KGEN_MACHINE):
            try:
                inc = KgenConfigParser(allow_no_value=True)
                inc.read(path)
                cmd = inc.get('shell', 'command')
                out, err, retcode = run_shcmd(cmd)
                if retcode == 0:
                    if inc.has_option('shell', 'startswith'):
                        if any( out.startswith(s.strip()) for s in inc.get('shell', 'startswith').split(',') ):
                            break
                    elif inc.has_option('shell', 'pattern'):
                        if any( re.match(p.strip(), out) for p in inc.get('shell', 'pattern').split(',') ):
                            break
                inc = None
            except:
                inc = None

        if inc is None:
            inc = KgenConfigParser(allow_no_value=True)
            inc.read('%s/generic_Linux'%KGEN_MACHINE)

        return inc


    def read_machinefile(self, machine, prerun):

        # populate contents of the machine file
        try:
            inc = machine['inc']
            if inc.has_section('general'):
                if inc.has_option('general', 'name') and inc.get('general', 'name'):
                    machine['general']['name'] = inc.get('general', 'name')
                if inc.has_option('general', 'id') and inc.get('general', 'id'):
                    machine['general']['id'] = inc.get('general', 'id')
            if inc.has_section('variable'):
                if inc.has_option('variable', 'prerun_build') and inc.get('variable', 'prerun_build'):
                    prerun['build'] = inc.get('variable', 'prerun_build')
                if inc.has_option('variable', 'prerun_run') and inc.get('variable', 'prerun_run'):
                    prerun['run'] = inc.get('variable', 'prerun_run')
                if inc.has_option('variable', 'prerun_kernel_build') and inc.get('variable', 'prerun_kernel_build'):
                    prerun['kernel_build'] = inc.get('variable', 'prerun_kernel_build')
                if inc.has_option('variable', 'prerun_kernel_run') and inc.get('variable', 'prerun_kernel_run'):
                    prerun['kernel_run'] = inc.get('variable', 'prerun_kernel_run')
                if inc.has_option('variable', 'work_directory') and inc.get('variable', 'work_directory'):
                    machine['variable']['work_directory'] = os.path.expandvars(inc.get('variable', 'work_directory'))
            if inc.has_section('compiler'):
                if inc.has_option('compiler', 'gnu') and inc.get('compiler', 'gnu'):
                    machine['compiler']['gnu'] = inc.get('compiler', 'gnu')
                if inc.has_option('compiler', 'intel') and inc.get('compiler', 'intel'):
                    machine['compiler']['intel'] = inc.get('compiler', 'intel')
                if inc.has_option('compiler', 'pgi') and inc.get('compiler', 'pgi'):
                    machine['compiler']['pgi'] = inc.get('compiler', 'pgi')
        except:
            pass

def handle_include(mpifdir, lines, config):

    insert_lines = []
    for i, line in enumerate(lines):
        match = re.match(r'^\s*include\s*("[^"]+"|\'[^\']+\')\s*\Z', line, re.I)
        if match:
            include_dirs = [mpifdir]+config["include"]['path']
            filename = match.group(1)[1:-1].strip()
            path = filename
            for incl_dir in include_dirs:
                path = os.path.join(incl_dir, filename)
                if os.path.exists(path):
                    break
            if os.path.isfile(path):
                with open(path, 'r') as f:
                    included_lines = f.read()
                    insert_lines.extend(handle_include(mpifdir, included_lines.split('\n'), config))
            else:
                raise UserException('Can not find %s in include paths.'%path)
        else:
            insert_lines.append(line)

    return insert_lines



# gen_core_utils.py

from collections import OrderedDict
from fortlab.resolver import typedecl_statements, block_statements

shared_objects = OrderedDict()

kernel_verify_kgenutils = []
kernel_verify_contains = []

VERIFY_PBLOCK_USE_PART = 'VPBUP'
VERIFY_PBLOCK_DECL_PART = 'VPBDP'
VERIFY_PBLOCK_EXEC_PART ='VPBEP'
VERIFY_PBLOCK_CONTAINS_PART = 'VPBCP'
VERIFY_PBLOCK_SUBP_PART = 'VPBSP'
VERIFY_PBLOCK_INIT = 'VPBI'
VERIFY_PBLOCK_EXTERNS = 'VPBE'
VERIFY_PBLOCK_LOCALS ='VPBL'

vprefix = 'kv'
MAXLEN_SUBPNAME = 40

def get_ancestor_name(stmt, generation):
    assert stmt and hasattr(stmt, 'parent'), 'Given stmt does not have parent attribute.'
    assert isinstance(generation, int), 'Not integer type of generation.'

    ancestor = stmt.ancestors()[generation]
    if hasattr(ancestor, 'name'):
        return ancestor.name 
    else: return ''

def get_topname(stmt):
    return get_ancestor_name(stmt, 0)

def get_parentname(stmt):
    return get_ancestor_name(stmt, -1)

def get_dtype_subpname(typestmt):
    if not hasattr(get_dtype_subpname, 'kgen_subpname_cache'):
        get_dtype_subpname.kgen_subpname_cache = OrderedDict()

    assert typestmt, 'None type of typestmt'

    subpname = '%s_%s'%(get_topname(typestmt), typestmt.name)

    if len(subpname)<MAXLEN_SUBPNAME:
        return subpname
    else:
        if subpname in get_dtype_subpname.kgen_subpname_cache:
            return 'kgen_%s_typesubp%d'%(get_topname(typestmt), get_dtype_subpname.kgen_subpname_cache[subpname])
        else:
            subpindex = len(get_dtype_subpname.kgen_subpname_cache)
            get_dtype_subpname.kgen_subpname_cache[subpname] = subpindex
            return 'kgen_%s_typesubp%d'%(get_topname(typestmt), subpindex)

def get_module_verifyname(modstmt):
    if modstmt is None: return
    return '%s_externs_%s'%(vprefix, modstmt.name)

def get_typedecl_subpname(stmt, entity_name):
    if not hasattr(get_typedecl_subpname, 'kgen_subpname_cache'):
        get_typedecl_subpname.kgen_subpname_cache = OrderedDict()

    assert isinstance(stmt, typedecl_statements.TypeDeclarationStatement), 'None type of typedecl stmt'
    assert entity_name, 'No entity name is provided.'

    var = stmt.get_variable(entity_name)
    if var is None: return 'Unknown_name'

    prefix = [ get_parentname(stmt), stmt.name ] + list(stmt.selector)
    l = []
    if var.is_array(): l.append('dim%d'%var.rank)
    if var.is_pointer(): l.append('ptr')

    subpname = '_'.join(prefix+l)
    if len(subpname)<MAXLEN_SUBPNAME:
        return '_'.join(prefix+l)
    else:
        if subpname in get_typedecl_subpname.kgen_subpname_cache:
            return 'kgen_%s_subp%d'%(get_parentname(stmt), get_typedecl_subpname.kgen_subpname_cache[subpname])
        else:
            subpindex = len(get_typedecl_subpname.kgen_subpname_cache)
            get_typedecl_subpname.kgen_subpname_cache[subpname] = subpindex
            return 'kgen_%s_subp%d'%(get_parentname(stmt), subpindex)

def get_dtype_verifyname(typestmt):
    if typestmt is None: return
    subpname = get_dtype_subpname(typestmt)
    if subpname: return '%s_%s'%(vprefix, subpname)

def get_typedecl_verifyname(typestmt, entity_name):
    if typestmt is None: return
    subpname = get_typedecl_subpname(typestmt, entity_name)
    if subpname: return '%s_%s'%(vprefix, subpname)

def process_spec_stmts(stmt):
    if not stmt: return
    if not hasattr(stmt, 'spec_stmts'): return

    for spec_stmt in stmt.spec_stmts:
        node = spec_stmt.genkpair
        if not node: continue
        if not node.kgen_isvalid: continue
        if not hasattr(spec_stmt, 'geninfo') or len(spec_stmt.geninfo)==0: continue

        if hasattr(spec_stmt, 'items'):
            new_items = []
            unames = list(set([ uname.firstpartname() for uname, req in KGGenType.get_state(spec_stmt.geninfo) ]))
            for item in spec_stmt.items:
                if any(item.startswith(uname) for uname in unames):
                    new_items.append(item)
            node.new_items = new_items
            node.kgen_use_tokgen = True
        else:
            pass
            # maybe specific handling per classes

def is_excluded(ename, stmt):
    if hasattr(stmt, 'exclude_names'):
        for name, actions in stmt.exclude_names.items():
            if ename==name: return True
    return False

def is_remove_state(ename, stmt):
    if hasattr(stmt, 'exclude_names'):
        for name, actions in stmt.exclude_names.items():
            if ename==name and 'remove_state' in actions:
                return True
    return False

def is_param_zero(length, stmt):

    if hasattr(stmt, 'unknowns'):
        for uname, req in stmt.unknowns.items():
            if uname.firstpartname()==length and len(req.res_stmts)>0:
                res_stmt = req.res_stmts[0]
                if isinstance(res_stmt, typedecl_statements.Integer) and 'parameter' in res_stmt.attrspec:
                    for decl in res_stmt.entity_decls:
                        vname, value = decl.split('=')            
                        if vname.strip()==length:
                            try:
                                intlen = int(value)
                                if intlen == 0: return True
                            except:
                                if stmt is res_stmt:
                                    print ('WARNING: recursive size check of following statement\n%s'%str(res_stmt))                                
                                else:
                                    return is_param_zero(length, res_stmt)
    return False

def is_zero_array(var, stmt):
    # Temporary turn off
    #return False
    if var.is_explicit_shape_array():
        for length in var.shape:
            try:
                intlen = int(length)
                if intlen==0: return True
            except:
                if is_param_zero(length, stmt):
                    return True
    return False

def check_class_derived(stmt):

    if not stmt.is_class(): return False

    # check if the name of kind in unknows, and the last res of the unknown is derivied type
    dtypename = stmt.selector[1]
    if hasattr(stmt, 'unknowns'):
        for uname, req in stmt.unknowns.items():
            if uname.firstpartname()==dtypename and len(req.res_stmts)>0:
                if isinstance(req.res_stmts[0], (block_statements.Type, block_statements.TypeDecl)):
                    return True
    return False



# kgen_analyze.py


from fortlab.kgutils import KGName, ProgramException, UserException, traverse
from fortlab.resolver.kgparse import KGGenType, SrcFile, ResState
from fortlab.resolver.Fortran2003 import Name, Call_Stmt, Function_Reference, Part_Ref, Interface_Stmt, Actual_Arg_Spec_List, \
    Section_Subscript_List, Actual_Arg_Spec, Structure_Constructor_2, Proc_Component_Ref, Add_Operand
from collections import OrderedDict
from fortlab.resolver.typedecl_statements import TypeDeclarationStatement
from fortlab.resolver.block_statements import SubProgramStatement, Associate

def update_state_info(parent, config):

    def get_nodes(node, bag, depth):
        from fortlab.resolver.Fortran2003 import Name
        if isinstance(node, Name) and node.string==bag['name'] and not node.parent in bag:
            anc = [node]
            while hasattr(node, 'parent'):
                anc.insert(0, node.parent)
                node = node.parent
            bag['lineage'].append(anc)

    if hasattr(parent, 'content'):
        for stmt in parent.content:
            if isinstance(stmt, TypeDeclarationStatement) and \
                "parameter" not in stmt.attrspec and hasattr(stmt, 'geninfo') and \
                any(len(v)>0 for v in stmt.geninfo.values()):
                for uname, req in KGGenType.get_state_in(stmt.geninfo):
                    if KGGenType.has_uname_out(uname, stmt.geninfo): continue
                    # select names for searching
                    respairs = []
                    if req.originator in config["callsite"]['stmts']:
                        respairs.append((uname, req.originator))
                    elif isinstance(req.originator, Associate):
                        if uname in req.originator.assoc_map:
                            for auname in req.originator.assoc_map[uname]:
                                for aruname, arreq in KGGenType.get_state_in(req.originator.geninfo):
                                    if auname == aruname:
                                        respairs.append((auname, arreq.originator))

                    if len(respairs) > 0:
                        copied = False
                        for varname, org in respairs:
                            bag = {'name': varname.firstpartname(), 'lineage': [] }
                            traverse(org.f2003, get_nodes, bag)
                            for lineage in bag['lineage']:
                                for lidx, anc in enumerate(lineage):
                                    # get callname
                                    callname = None
                                    if anc.__class__ in [ Call_Stmt, Function_Reference ]:
                                        callname = anc.items[0].string
                                    elif anc.__class__ == Part_Ref:
                                        callname = anc.items[0].string
                                    elif anc.__class__ == Interface_Stmt:
                                        callname = anc.items[0].string

                                    # get caller and callee objects
                                    callobj = None
                                    subpobj = None
                                    if callname:
                                        for org_uname, org_req in org.unknowns.items():
                                            if org_uname.firstpartname()==callname:
                                                if isinstance(org_req.res_stmts[0], SubProgramStatement):
                                                    callobj = anc
                                                    subpobj = org_req.res_stmts[0]
                                                break
                                        
                                    # get argument index
                                    argidx = -1
                                    is_keyword = False
                                    if callobj and subpobj:
                                        if callobj.__class__ in [ Call_Stmt, Function_Reference ]:
                                            arglist = callobj.items[1]
                                            if arglist is None: pass
                                            elif isinstance(arglist, Actual_Arg_Spec):
                                                argobj = lineage[lidx+1]
                                                kword = argobj.items[0].string
                                                argidx = 0

                                                if kword in subpobj.args:
                                                    argidx = subpobj.args.index(kword)
                                                
                                            elif isinstance(arglist, Actual_Arg_Spec_List):
                                                argobj = lineage[lidx+2]
                                                if isinstance(argobj, Actual_Arg_Spec):
                                                    argidx = arglist.items.index(argobj)
                                                    kword = argobj.items[0].string
                                                    if kword in subpobj.args:
                                                        argidx = subpobj.args.index(kword)

                                                elif isinstance(argobj, (Name, Proc_Component_Ref,
                                                                Add_Operand, Part_Ref)):
                                                    argidx = arglist.items.index(argobj)
   
                                                else:
                                                    import pdb; pdb.set_trace()
                                                    raise Exception("Not implemented to handle: %s", argobj.__class__.__name__)
                                            else:
                                                argidx = 0
                                        elif anc.__class__ == Part_Ref:
                                            arglist = callobj.items[1]
                                            if arglist is None: pass
                                            elif isinstance(arglist, Structure_Constructor_2):
                                                argobj = lineage[lidx+1]
                                                kword = argobj.items[0].string
                                                argidx = subpobj.args.index(kword)
                                            elif isinstance(arglist, Section_Subscript_List):
                                                argobj = lineage[lidx+2]
                                                argidx = arglist.items.index(argobj)
                                                if isinstance(argobj, Structure_Constructor_2):
                                                    kword = argobj.items[0].string
                                                    argidx = subpobj.args.index(kword)
                                            else:
                                                argidx = 0
                                        elif anc.__class__ == Interface_Stmt:
                                            raise Exception("Interface Statement is not expected.")

                                    # get intent
                                    if argidx>=0:
                                        if argidx >= len(subpobj.args):
                                            raise Exception("argument index exceeds the length of arg. list: %d >= %d",
                                                argidx, len(subpobj.args))
                                        argname = subpobj.args[argidx]
                                        var = subpobj.a.variables[subpobj.args[argidx]]
                                        if var.is_intent_out() or var.is_intent_inout():
                                            req.gentype = KGGenType.STATE_OUT
                                            stmt.add_geninfo(uname, req)
                                            copied = True
                                            break
                                if copied: break
                            if copied: break

    if hasattr(parent, 'parent'):
        update_state_info(parent.parent, config)


def analyze(config):

    analyze_callsite(config)

def analyze_callsite(config):
    from fortlab.resolver.block_statements import EndStatement, Subroutine, Function, Interface
    from fortlab.resolver.statements import SpecificBinding
    from fortlab.resolver.kgsearch import f2003_search_unknowns

    # read source file that contains callsite stmt
    cs_file = SrcFile(config["callsite"]['filepath'])

    #process_directive(cs_file.tree)

    if len(config["callsite"]['stmts'])==0:
        raise UserException('Can not find callsite')

    # ancestors of callsite stmt
    ancs = config["callsite"]['stmts'][0].ancestors()

    # add geninfo for ancestors
    prevstmt = config["callsite"]['stmts'][0]
    prevname = None

    for anc in reversed(ancs):
        if not hasattr(anc, 'geninfo'):
            anc.geninfo = OrderedDict()
        if len(anc.content)>0 and isinstance(anc.content[-1], EndStatement) and \
            not hasattr(anc.content[-1], 'geninfo'):
            anc.content[-1].geninfo = OrderedDict()

        if prevname:
            dummy_req = ResState(KGGenType.STATE_IN, KGName(prevname), None, [anc])
            dummy_req.res_stmts = [ prevstmt ]
            anc.check_spec_stmts(dummy_req.uname, dummy_req, config)

        if hasattr(anc, 'name'): prevname = anc.name
        else: prevname = None
        prevstmt = anc

    # populate parent block parameters
    config["parentblock"]['stmt'] = ancs[-1]

    # populate top block parameters
    config["topblock"]['stmt'] = ancs[0]

    for cs_stmt in config["callsite"]['stmts']:
        #resolve cs_stmt
        f2003_search_unknowns(cs_stmt, cs_stmt.f2003, config)
        for uname, req in cs_stmt.unknowns.items():
            cs_stmt.resolve(req, config)
            if not req.res_stmts:
                raise ProgramException('Resolution fail.')


    # update state info of callsite and its upper blocks
    update_state_info(config["parentblock"]['stmt'], config)

    # update state info of modules
    for modname, moddict in config["modules"].items():
        modstmt = moddict['stmt']
        if modstmt != config["topblock"]['stmt']:
            update_state_info(moddict['stmt'], config)



"""
Provides get_source_info(<filename>) function to determine the format
(free|fixed|strict|pyf) of a Fortran file.

-----
Permission to use, modify, and distribute this software is given under the
terms of the NumPy License. See http://scipy.org.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Author: Pearu Peterson <pearu@cens.ioc.ee>
Created: May 2006
-----
"""

__all__ = ['get_source_info', 'get_source_info_str']

import re
import os
import sys

_has_f_extension = re.compile(r'.*[.](for|ftn|f77|f)\Z',re.I).match
_has_f_header = re.compile(r'-[*]-\s*(fortran|f77)\s*-[*]-',re.I).search
_has_f90_header = re.compile(r'-[*]-\s*f90\s*-[*]-',re.I).search
_has_f03_header = re.compile(r'-[*]-\s*f03\s*-[*]-',re.I).search
_has_f08_header = re.compile(r'-[*]-\s*f08\s*-[*]-',re.I).search
_has_free_header = re.compile(r'-[*]-\s*(f90|f95|f03|f08)\s*-[*]-',re.I).search
_has_fix_header = re.compile(r'-[*]-\s*fix\s*-[*]-',re.I).search
_has_pyf_header = re.compile(r'-[*]-\s*pyf\s*-[*]-',re.I).search
_free_format_start = re.compile(r'[^c*!]\s*[^\s\d\t]',re.I).match

def get_source_info(filename):
    """
    Determine if fortran file is
      - in fix format and contains Fortran 77 code    -> return False, True
      - in fix format and contains Fortran 90,95,2003,2008 code    -> return False, False
      - in free format and contains Fortran 90,95,2003,2008 code   -> return True, False
      - in free format and contains signatures (.pyf) -> return True, True
    """
    base,ext = os.path.splitext(filename)
    if ext=='.pyf':
        return True, True
    isfree = False
    isstrict = False
    f = open(filename,'r')
    firstline = f.readline()
    f.close()
    if _has_f_header(firstline): return False, True
    if _has_fix_header(firstline): return False, False
    if _has_free_header(firstline): return True, False
    if _has_pyf_header(firstline): return True, True
    if _has_f_extension(filename) and \
       not (_has_free_header(firstline) or _has_fix_header(firstline)):
        isstrict = True
    elif is_free_format(filename) and not _has_fix_header(firstline):
        isfree = True
    return isfree,isstrict

def get_source_info_str(code):
    """
    Determine the format of Fortran code.
    See get_source_info() for the meaning of return values.
    """
    firstline = code.lstrip().split('\n',1)[0]
    if _has_f_header(firstline): return False, True
    if _has_fix_header(firstline): return False, False
    if _has_free_header(firstline): return True, False
    if _has_pyf_header(firstline): return True, True
    n = 10000 # the number of non-comment lines to scan for hints
    lines = code.splitlines()
    isfree = False
    while n>0 and lines:
        line = lines.pop(0).rstrip()
        if line and line[0]!='!':
            n -= 1
            if line[0]!='\t' and _free_format_start(line[:5]) or line[-1:]=='&':
                isfree = True
                break
        
    return isfree, False # assume free f90 or fix f90.

def is_free_format(file):
    """Check if file is in free format Fortran."""
    # f90-2008 allows both fixed and free format, assuming fixed
    # unless signs of free format are detected.
    isfree = False
    f = open(file,'r')
    line = f.readline()
    n = 10000 # the number of non-comment lines to scan for hints
    if _has_f_header(line):
        n = 0
    elif _has_free_header(line):
        n = 0
        isfree = True
    while n>0 and line:
        line = line.rstrip()
        if line and line[0]!='!':
            n -= 1
            if line[0]!='\t' and _free_format_start(line[:5]) or line[-1:]=='&':
                isfree = True
                break
        line = f.readline()
    f.close()
    return isfree

def simple_main():
    for filename in sys.argv[1:]:
        isfree, isstrict = get_source_info(filename)
        print('%s: isfree=%s, isstrict=%s'  % (filename, isfree, isstrict))

if __name__ == '__main__':
    simple_main()



#!/usr/bin/env python
"""Provides FortranParser.
"""
#Author: Pearu Peterson <pearu@cens.ioc.ee>
#Created: May 2006

__autodoc__ = ['FortranParser']
__all__ = ['FortranParser']

import re
import sys
import traceback
#from numpy.distutils.misc_util import yellow_text, red_text # KGEN deletion

from fortlab.resolver.readfortran import FortranFileReader, FortranStringReader
from fortlab.resolver.block_statements import BeginSource
from fortlab.resolver.utils import AnalyzeError
from fortlab.kgutils import logger

#import logging
#logger = logging.getLogger('kgen')

class FortranParser(object):
    """Parser of FortranReader structure.
    
    Use .parse() method for parsing, parsing result is saved in .block attribute.
    """
    cache = {}

    def __init__(self, reader, ignore_comments=True):
        self.reader = reader
        if reader.id in self.cache:
            parser = self.cache[reader.id]
            self.block = parser.block
            self.is_analyzed = parser.is_analyzed
            logger.debug('using cached %s' % (reader.id))
            # self.block.show_message('using cached %s' % (reader.id))
        else:
            self.cache[reader.id] = self
            self.block = None
            self.is_analyzed = False
        self.ignore_comments = ignore_comments
        return

    def get_item(self):
        try:
            item = self.reader.next(ignore_comments = self.ignore_comments)
            return item
        except StopIteration:
            pass
        return

    def put_item(self, item):
        self.reader.fifo_item.insert(0, item)
        return

    def parse(self):
        if self.block is not None:
            return
        try:
            block = self.block = BeginSource(self)
        except KeyboardInterrupt:
            raise
        except Exception as err:
            reader = self.reader
            while reader is not None:
                message = reader.format_message('FATAL ERROR',
                                                'while processing line',
                                                reader.linecount, reader.linecount)
                logger.critical(message)
                # reader.show_message(message, sys.stderr)
                reader = reader.reader
            logger.debug(''.join(('Traceback\n',''.join( traceback.format_stack() ))))
            #traceback.print_exc(file=sys.stderr)
            #logger.critical(red_text('STOPPED PARSING')) # KGEN deletion
            logger.critical('STOPPED PARSING: %s' % str(err)) # KGEN addition
            # self.reader.show_message(red_text('STOPPED PARSING'), sys.stderr)
            return
        return

    def analyze(self):
        if self.is_analyzed:
            return
        if self.block is None:
            logger.debug('Nothing to analyze.')
            # self.reader.show_message('Nothing to analyze.')
            return

        try:
            self.block.analyze()
        except AnalyzeError:
            pass
        # except Exception, msg:
            # import pdb; pdb.set_trace()
            # if str(msg) != '123454321':
                # #print self.block
                # logger.debug(''.join(('Traceback\n',''.join( traceback.format_stack() ))))
                # logger.critical(red_text('FATAL ERROR: STOPPED ANALYSING %r CONTENT' % (self.reader.source) ))
                # # self.reader.show_message(red_text('FATAL ERROR: STOPPED ANALYSING %r CONTENT' % (self.reader.source) ), sys.stderr)
                # sys.exit(123454321)
            # return
        self.is_analyzed = True
        return

def test_pyf():
    string = """
python module foo
  interface tere
    subroutine bar
    real r
    end subroutine bar
  end interface tere
end python module foo
"""
    reader = FortranStringReader(string, True, True)
    parser = FortranParser(reader)
    block = parser.parse()
    print(block)

def test_free90():
    string = """
module foo

   subroutine bar
    real r
    if ( pc_get_lun() .ne. 6) &
    write ( pc_get_lun(), '( &
    & /, a, /, " p=", i4, " stopping c_flag=", a, &
    & /, " print unit=", i8)') &
    trim(title), pcpsx_i_pel(), trim(c_flag), pc_get_lun()
    if (.true.) then
      call smth
    end if
    aaa : if (.false.) then
    else if (a) then aaa
    else aaa
    end if aaa
    hey = 1
    end subroutine bar
    abstract interface

    end interface

end module foo
"""
    reader = FortranStringReader(string, True, False)
    parser = FortranParser(reader)
    block = parser.parse()
    print(block)

def test_f77():
    string = """\
      program foo
      a = 3
      end
      subroutine bar
      end
      pure function foo(a)
      end
      pure real*4 recursive function bar()
      end
"""
    reader = FortranStringReader(string, False, True)
    parser = FortranParser(reader)
    block = parser.parse()
    print(block)

def simple_main():
    import sys
    if not sys.argv[1:]:
        return parse_all_f()
    for filename in sys.argv[1:]:
        reader = FortranFileReader(filename)
        print(yellow_text('Processing '+filename+' (mode=%r)' % (reader.mode)))
        parser = FortranParser(reader)
        parser.parse()
        parser.analyze()
        print(parser.block.torepr(4))
        #print parser.block

def profile_main():
    import hotshot, hotshot.stats
    prof = hotshot.Profile("_parsefortran.prof")
    prof.runcall(simple_main)
    prof.close()
    stats = hotshot.stats.load("_parsefortran.prof")
    stats.strip_dirs()
    stats.sort_stats('time', 'calls')
    stats.print_stats(30)

def parse_all_f():
    for filename in open('opt_all_f.txt'):
        filename = filename.strip()
        reader = FortranFileReader(filename)
        print(yellow_text('Processing '+filename+' (mode=%r)' % (reader.mode)))
        parser = FortranParser(reader)
        block = parser.parse()
        print(block)

if __name__ == "__main__":
    #test_f77()
    #test_free90()
    #test_pyf()
    simple_main()
    #profile_main()
    #parse_all_f()



# kgen_extra.py

kgen_file_header = \
"""
! KGEN-generated Fortran source file
!
! Filename    : %s
! Generated at: %s
! KGEN version: %s

"""

kgen_subprograms = \
"""FUNCTION kgen_get_newunit() RESULT(new_unit)
   INTEGER, PARAMETER :: UNIT_MIN=100, UNIT_MAX=1000000
   LOGICAL :: is_opened
   INTEGER :: nunit, new_unit, counter

   new_unit = -1
   DO counter=UNIT_MIN, UNIT_MAX
       inquire(UNIT=counter, OPENED=is_opened)
       IF (.NOT. is_opened) THEN
           new_unit = counter
           EXIT
       END IF
   END DO
END FUNCTION

SUBROUTINE kgen_error_stop( msg )
    IMPLICIT NONE
    CHARACTER(LEN=*), INTENT(IN) :: msg

    WRITE (*,*) msg
    STOP 1
END SUBROUTINE """

kgen_print_counter = \
"""SUBROUTINE kgen_print_counter(counter)
    INTEGER, INTENT(IN) :: counter
    PRINT *, "KGEN writes input state variables at count = ", counter
END SUBROUTINE

SUBROUTINE kgen_print_mpirank_counter(rank, counter)
    INTEGER, INTENT(IN) :: rank, counter
    PRINT *, "KGEN writes input state variables at count = ", counter, " on mpirank = ", rank
END SUBROUTINE"""


kgen_verify_intrinsic_checkpart = \
"""check_status%%numTotal = check_status%%numTotal + 1
IF ( var %s ref_var ) THEN
    check_status%%numIdentical = check_status%%numIdentical + 1
    if(kgen_verboseLevel == 3) then
        WRITE(*,*)
        WRITE(*,*) trim(adjustl(varname)), " is IDENTICAL( ", var, " )."
    endif
ELSE
    if(kgen_verboseLevel > 0) then
        WRITE(*,*)
        WRITE(*,*) trim(adjustl(varname)), " is NOT IDENTICAL."
        if(kgen_verboseLevel == 3) then
            WRITE(*,*) "KERNEL: ", var
            WRITE(*,*) "REF.  : ", ref_var
        end if
    end if
    check_status%%numOutTol = check_status%%numOutTol + 1
END IF"""

kgen_verify_numeric_array = \
"""check_status%%numTotal = check_status%%numTotal + 1
IF ( ALL( var %(eqtest)s ref_var ) ) THEN

    check_status%%numIdentical = check_status%%numIdentical + 1            
    if(kgen_verboseLevel == 3) then
        WRITE(*,*)
        WRITE(*,*) "All elements of ", trim(adjustl(varname)), " are IDENTICAL."
        !WRITE(*,*) "KERNEL: ", var
        !WRITE(*,*) "REF.  : ", ref_var
        IF ( ALL( var == 0 ) ) THEN
            if(kgen_verboseLevel == 3) then
                WRITE(*,*) "All values are zero."
            end if
        END IF
    end if
ELSE
    allocate(temp(%(allocshape)s))
    allocate(temp2(%(allocshape)s))

    n = count(var/=ref_var)
    where(abs(ref_var) > kgen_minvalue)
        temp  = ((var-ref_var)/ref_var)**2
        temp2 = (var-ref_var)**2
    elsewhere
        temp  = (var-ref_var)**2
        temp2 = temp
    endwhere
    nrmsdiff = sqrt(sum(temp)/real(n))
    rmsdiff = sqrt(sum(temp2)/real(n))

    if (nrmsdiff > kgen_tolerance) then
        check_status%%numOutTol = check_status%%numOutTol+1
    else
        check_status%%numInTol = check_status%%numInTol+1
    endif

    deallocate(temp,temp2)
END IF"""

kgen_verify_nonreal_array = \
"""check_status%%numTotal = check_status%%numTotal + 1
IF ( ALL( var %(eqtest)s ref_var ) ) THEN

    check_status%%numIdentical = check_status%%numIdentical + 1            
    if(kgen_verboseLevel == 3) then
        WRITE(*,*)
        WRITE(*,*) "All elements of ", trim(adjustl(varname)), " are IDENTICAL."
        !WRITE(*,*) "KERNEL: ", var
        !WRITE(*,*) "REF.  : ", ref_var
        IF ( ALL( var == 0 ) ) THEN
                WRITE(*,*) "All values are zero."
        END IF
    end if
ELSE
    if(kgen_verboseLevel > 0) then
        WRITE(*,*)
        WRITE(*,*) trim(adjustl(varname)), " is NOT IDENTICAL."
        WRITE(*,*) count( var /= ref_var), " of ", size( var ), " elements are different."
    end if

    check_status%%numOutTol = check_status%%numOutTol+1
END IF"""

kgen_utils_file_head = \
"""
INTEGER, PARAMETER :: kgen_dp = selected_real_kind(15, 307)
INTEGER, PARAMETER :: CHECK_IDENTICAL = 1
INTEGER, PARAMETER :: CHECK_IN_TOL = 2
INTEGER, PARAMETER :: CHECK_OUT_TOL = 3

REAL(kind=kgen_dp) :: kgen_tolerance = 1.0D-15, kgen_minvalue = 1.0D-15
INTEGER :: kgen_verboselevel = 1

interface kgen_tostr
    module procedure kgen_tostr_args1
    module procedure kgen_tostr_args2
    module procedure kgen_tostr_args3
    module procedure kgen_tostr_args4
    module procedure kgen_tostr_args5
    module procedure kgen_tostr_args6
end interface

! PERTURB: add following interface
interface kgen_perturb_real
    module procedure kgen_perturb_real4_dim1
    module procedure kgen_perturb_real4_dim2
    module procedure kgen_perturb_real4_dim3
    module procedure kgen_perturb_real8_dim1
    module procedure kgen_perturb_real8_dim2
    module procedure kgen_perturb_real8_dim3
end interface

type check_t
    logical :: Passed
    integer :: numOutTol
    integer :: numTotal
    integer :: numIdentical
    integer :: numInTol
    integer :: rank
end type check_t

public kgen_dp, check_t, kgen_init_verify, kgen_init_check, kgen_tolerance
public kgen_minvalue, kgen_verboselevel, kgen_print_check, kgen_perturb_real
public CHECK_NOT_CHECKED, CHECK_IDENTICAL, CHECK_IN_TOL, CHECK_OUT_TOL
public kgen_get_newunit, kgen_error_stop
"""

kgen_utils_array_sumcheck = \
"""
subroutine kgen_array_sumcheck(varname, sum1, sum2, finish)
    character(*), intent(in) :: varname
    real(kind=8), intent(in) :: sum1, sum2
    real(kind=8), parameter  :: max_rel_diff = 1.E-10
    real(kind=8)  :: diff, rel_diff
    logical, intent(in), optional :: finish
    logical checkresult

    if ( sum1 == sum2 ) then
        checkresult = .TRUE.
    else
        checkresult = .FALSE.

        diff = ABS(sum2 - sum1)

        if ( .NOT. (sum1 == 0._8) ) then

            rel_diff = ABS(diff / sum1)
            if ( rel_diff > max_rel_diff ) then

                print *, ''
                print *, 'SUM of array, "', varname, '", is different.'
                print *, 'From file : ', sum1
                print *, 'From array: ', sum2
                print *, 'Difference: ', diff
                print *, 'Normalized difference: ', rel_diff

                if ( present(finish) .AND. finish ) then
                    stop
                end if
            end if
        else
            print *, ''
            print *, 'SUM of array, "', varname, '", is different.'
            print *, 'From file : ', sum1
            print *, 'From array: ', sum2
            print *, 'Difference: ', diff

            if ( present(finish) .AND. finish ) then
                stop
            end if
        end if
    end if
end subroutine
"""

kgen_utils_file_tostr = \
"""
function kgen_tostr_args1(idx1) result(tostr)
    integer, intent(in) :: idx1
    character(len=64) :: str_idx1
    character(len=64) :: tostr

    write(str_idx1, *) idx1
    tostr = trim(adjustl(str_idx1))
end function

function kgen_tostr_args2(idx1, idx2) result(tostr)
    integer, intent(in) :: idx1, idx2
    character(len=64) :: str_idx1, str_idx2
    character(len=128) :: tostr

    write(str_idx1, *) idx1
    write(str_idx2, *) idx2
    tostr = trim(adjustl(str_idx1)) // ", " // trim(adjustl(str_idx2))
end function

function kgen_tostr_args3(idx1, idx2, idx3) result(tostr)
    integer, intent(in) :: idx1, idx2, idx3
    character(len=64) :: str_idx1, str_idx2, str_idx3
    character(len=192) :: tostr

    write(str_idx1, *) idx1
    write(str_idx2, *) idx2
    write(str_idx3, *) idx3
    tostr = trim(adjustl(str_idx1)) // ", " // trim(adjustl(str_idx2)) &
        // ", " // trim(adjustl(str_idx3))
end function

function kgen_tostr_args4(idx1, idx2, idx3, idx4) result(tostr)
    integer, intent(in) :: idx1, idx2, idx3, idx4
    character(len=64) :: str_idx1, str_idx2, str_idx3, str_idx4
    character(len=256) :: tostr

    write(str_idx1, *) idx1
    write(str_idx2, *) idx2
    write(str_idx3, *) idx3
    write(str_idx4, *) idx4
    tostr = trim(adjustl(str_idx1)) // ", " // trim(adjustl(str_idx2)) &
        // ", " // trim(adjustl(str_idx3)) // ", " // trim(adjustl(str_idx4))
end function

function kgen_tostr_args5(idx1, idx2, idx3, idx4, idx5) result(tostr)
    integer, intent(in) :: idx1, idx2, idx3, idx4, idx5
    character(len=64) :: str_idx1, str_idx2, str_idx3, str_idx4, str_idx5
    character(len=320) :: tostr

    write(str_idx1, *) idx1
    write(str_idx2, *) idx2
    write(str_idx3, *) idx3
    write(str_idx4, *) idx4
    write(str_idx5, *) idx5
    tostr = trim(adjustl(str_idx1)) // ", " // trim(adjustl(str_idx2)) &
        // ", " // trim(adjustl(str_idx3)) // ", " // trim(adjustl(str_idx4)) &
        // ", " // trim(adjustl(str_idx5))
end function

function kgen_tostr_args6(idx1, idx2, idx3, idx4, idx5, idx6) result(tostr)
    integer, intent(in) :: idx1, idx2, idx3, idx4, idx5, idx6
    character(len=64) :: str_idx1, str_idx2, str_idx3, str_idx4, str_idx5, str_idx6
    character(len=384) :: tostr

    write(str_idx1, *) idx1
    write(str_idx2, *) idx2
    write(str_idx3, *) idx3
    write(str_idx4, *) idx4
    write(str_idx5, *) idx5
    write(str_idx6, *) idx6
    tostr = trim(adjustl(str_idx1)) // ", " // trim(adjustl(str_idx2)) &
        // ", " // trim(adjustl(str_idx3)) // ", " // trim(adjustl(str_idx4)) &
        // ", " // trim(adjustl(str_idx5)) // ", " // trim(adjustl(str_idx6))
end function
"""

kgen_utils_file_checksubr = \
"""
subroutine kgen_perturb_real4_dim1(var, pertlim)
    real*4, intent(inout), dimension(:) :: var
    real*4, intent(in) :: pertlim
    integer, allocatable :: rndm_seed(:)
    integer :: rndm_seed_sz
    real*4 :: pertval
    integer :: idx1

    call random_seed(size=rndm_seed_sz)
    allocate(rndm_seed(rndm_seed_sz))
    rndm_seed = 121869
    call random_seed(put=rndm_seed)
    do idx1=1,size(var, dim=1)
        call random_number(pertval)
        pertval = 2.0_4*pertlim*(0.5_4 - pertval)
        var(idx1) = var(idx1)*(1.0_4 + pertval)
    end do
    deallocate(rndm_seed)
end subroutine

subroutine kgen_perturb_real4_dim2(var, pertlim)
    real*4, intent(inout), dimension(:,:) :: var
    real*4, intent(in) :: pertlim
    integer, allocatable :: rndm_seed(:)
    integer :: rndm_seed_sz
    real*4 :: pertval
    integer :: idx1,idx2

    call random_seed(size=rndm_seed_sz)
    allocate(rndm_seed(rndm_seed_sz))
    rndm_seed = 121869
    call random_seed(put=rndm_seed)
    do idx1=1,size(var, dim=1)
        do idx2=1,size(var, dim=2)
            call random_number(pertval)
            pertval = 2.0_4*pertlim*(0.5_4 - pertval)
            var(idx1,idx2) = var(idx1,idx2)*(1.0_4 + pertval)
        end do
    end do
    deallocate(rndm_seed)
end subroutine

subroutine kgen_perturb_real4_dim3(var, pertlim)
    real*4, intent(inout), dimension(:,:,:) :: var
    real*4, intent(in) :: pertlim
    integer, allocatable :: rndm_seed(:)
    integer :: rndm_seed_sz
    real*4 :: pertval
    integer :: idx1,idx2,idx3

    call random_seed(size=rndm_seed_sz)
    allocate(rndm_seed(rndm_seed_sz))
    rndm_seed = 121869
    call random_seed(put=rndm_seed)
    do idx1=1,size(var, dim=1)
        do idx2=1,size(var, dim=2)
            do idx3=1,size(var, dim=3)
                call random_number(pertval)
                pertval = 2.0_4*pertlim*(0.5_4 - pertval)
                var(idx1,idx2,idx3) = var(idx1,idx2,idx3)*(1.0_4 + pertval)
            end do
        end do
    end do
    deallocate(rndm_seed)
end subroutine

subroutine kgen_perturb_real8_dim1(var, pertlim)
    real*8, intent(inout), dimension(:) :: var
    real*8, intent(in) :: pertlim
    integer, allocatable :: rndm_seed(:)
    integer :: rndm_seed_sz
    real*8 :: pertval
    integer :: idx1

    call random_seed(size=rndm_seed_sz)
    allocate(rndm_seed(rndm_seed_sz))
    rndm_seed = 121869
    call random_seed(put=rndm_seed)
    do idx1=1,size(var, dim=1)
        call random_number(pertval)
        pertval = 2.0_8*pertlim*(0.5_8 - pertval)
        var(idx1) = var(idx1)*(1.0_8 + pertval)
    end do
    deallocate(rndm_seed)
end subroutine

subroutine kgen_perturb_real8_dim2(var, pertlim)
    real*8, intent(inout), dimension(:,:) :: var
    real*8, intent(in) :: pertlim
    integer, allocatable :: rndm_seed(:)
    integer :: rndm_seed_sz
    real*8 :: pertval
    integer :: idx1,idx2

    call random_seed(size=rndm_seed_sz)
    allocate(rndm_seed(rndm_seed_sz))
    rndm_seed = 121869
    call random_seed(put=rndm_seed)
    do idx1=1,size(var, dim=1)
        do idx2=1,size(var, dim=2)
            call random_number(pertval)
            pertval = 2.0_8*pertlim*(0.5_8 - pertval)
            var(idx1,idx2) = var(idx1,idx2)*(1.0_8 + pertval)
        end do
    end do
    deallocate(rndm_seed)
end subroutine

subroutine kgen_perturb_real8_dim3(var, pertlim)
    real*8, intent(inout), dimension(:,:,:) :: var
    real*8, intent(in) :: pertlim
    integer, allocatable :: rndm_seed(:)
    integer :: rndm_seed_sz
    real*8 :: pertval
    integer :: idx1,idx2,idx3

    call random_seed(size=rndm_seed_sz)
    allocate(rndm_seed(rndm_seed_sz))
    rndm_seed = 121869
    call random_seed(put=rndm_seed)
    do idx1=1,size(var, dim=1)
        do idx2=1,size(var, dim=2)
            do idx3=1,size(var, dim=3)
                call random_number(pertval)
                pertval = 2.0_8*pertlim*(0.5_8 - pertval)
                var(idx1,idx2,idx3) = var(idx1,idx2,idx3)*(1.0_8 + pertval)
            end do
        end do
    end do
    deallocate(rndm_seed)
end subroutine

subroutine kgen_init_verify(verboseLevel, tolerance, minValue)

  integer, intent(in), optional :: verboseLevel
  real(kind=kgen_dp), intent(in), optional :: tolerance
  real(kind=kgen_dp), intent(in), optional :: minValue

  if(present(verboseLevel)) then
     kgen_verboseLevel = verboseLevel
  end if

  if(present(tolerance)) then
      kgen_tolerance = tolerance
  end if

  if(present(minvalue)) then
      kgen_minvalue = minvalue
  end if

end subroutine kgen_init_verify

subroutine kgen_init_check(check, rank)

  type(check_t), intent(inout) :: check
  integer, intent(in), optional :: rank 

  check%Passed   = .TRUE.
  check%numOutTol = 0
  check%numInTol = 0
  check%numTotal = 0
  check%numIdentical = 0

  if(present(rank)) then                                                             
      check%rank = rank                                                              
  else                                                                               
      check%rank = 0                                                                 
  endif  

end subroutine kgen_init_check

subroutine kgen_print_check(kname, check)
   character(len=*) :: kname
   type(check_t), intent(in) ::  check

   write (*,*) TRIM(kname),': Tolerance for normalized RMS: ',kgen_tolerance
   !write (*,*) TRIM(kname),':',check%numFatal,'fatal errors,',check%numWarning,'warnings detected, and',check%numIdentical,'identical out of',check%numTotal,'variables checked'
   write (*,*) TRIM(kname),': Number of variables checked: ',check%numTotal
   write (*,*) TRIM(kname),': Number of Identical results: ',check%numIdentical
   write (*,*) TRIM(kname),': Number of variables within tolerance(not identical): ',check%numInTol
   write (*,*) TRIM(kname),': Number of variables out of tolerance: ', check%numOutTol

   if (check%numOutTol> 0) then
        write(*,*) TRIM(kname),': Verification FAILED'
   else
        write(*,*) TRIM(kname),': Verification PASSED'
   endif
end subroutine kgen_print_check
"""

kgen_get_newunit = \
"""
FUNCTION kgen_get_newunit() RESULT ( new_unit )
    INTEGER, PARAMETER :: UNIT_MIN=100, UNIT_MAX=1000000
    LOGICAL :: is_opened
    INTEGER :: nunit, new_unit, counter
    REAL :: r

    CALL RANDOM_SEED
    new_unit = -1
    DO counter=1, UNIT_MAX
        CALL RANDOM_NUMBER(r)
        nunit = INT(r*UNIT_MAX+UNIT_MIN)
        INQUIRE (UNIT=nunit, OPENED=is_opened)
        IF (.NOT. is_opened) THEN
            new_unit = nunit
            EXIT
        END IF
    END DO
END FUNCTION kgen_get_newunit
"""

kgen_error_stop = \
"""
SUBROUTINE kgen_error_stop( msg )
    IMPLICIT NONE
    CHARACTER(LEN=*), INTENT(IN) :: msg

    WRITE (*,*) msg
    STOP 1
END SUBROUTINE
"""

kgen_rankthread = \
"""
SUBROUTINE kgen_rankthreadinvoke( str, rank, thread, invoke )
    CHARACTER(*), INTENT(IN) :: str
    INTEGER, INTENT(OUT) :: rank, thread, invoke
    INTEGER :: pos1, pos2, i, e

    pos1 = 1

    rank = -1
    thread = -1
    invoke = -1

    DO
        pos2 = INDEX(str(pos1:), ".")
        IF (pos2 == 0) THEN
            READ(str(pos1:),*,IOSTAT=e) i
            IF ( e == 0 ) THEN
                rank = thread
                thread = invoke
                READ(str(pos1:), *) invoke
            END IF
            EXIT
        END IF

        READ(str(pos1:pos1+pos2-2),*,IOSTAT=e) i
        IF ( e == 0 ) THEN
            rank = thread
            thread = invoke
            READ(str(pos1:pos1+pos2-2), *) invoke
        END IF 

        pos1 = pos2+pos1
    END DO
END SUBROUTINE
"""

rdtsc = \
"""         .file   "rdtsc.s"
         .text
.globl rdtsc_
         .type   rdtsc_, @function
rdtsc_:
         rdtsc
         movl %eax,%ecx
         movl %edx,%eax
         shlq $32,%rax
         addq %rcx,%rax
         ret
         .size   rdtsc_, .-rdtsc_"""




Intrinsic_Procedures = [ \
# numeric functions \
'abs','aimag','aint','anint','ceiling','cmplx','conjg','dble','dim','dprod','floor','int','max','min','mod','modulo','nint','real','sign', \
# mathematical functions \
'acos','asin','atan','atan2','cos','cosh','exp','log','log10','sin','sinh','sqrt','tan','tanh', \
# character functions \
'achar','adjustl','adjustr','char','iachar','ichar','index','len_trim','lge','lgt','lle','llt','max','min','repeat','scan','trim','verify', \
# kind functions \
'kind','selected_char_kind','selected_int_kind','selected_real_kind', \
# miscellaneous type conversion functions \
'logical', 'transfer', \
# numeric inquiry functions \
'digits','epsilon','huge','maxexponent','minexponent','precision','radix','range','tiny', \
# array inquiry functions \
'lbound','shape','size','ubound', \
# other inquiry functions \
'allocated','associated','bit_size','extends_type_of','len','new_line','present','same_type_as', \
# bit manipulation procedures \
'btest','iand','ibclr','ibits','ibset','ieor','ior','ishft','ishftc','mvbits','not', \
# vector and matrix multiply functions \
'exponent','fraction','nearest','rrspacing','scale','set_exponent','spacing','dot_product','matmul', \
# array reduction functions \
'all' ,'any' ,'count' ,'maxval' ,'minval' ,'product' ,'sum', \
# array construction functions \
'cshift','eoshift','merge','pack','reshape','spread','transpose','unpack', \
# array location functions \
'maxloc','minloc', \
# null function \
'null', \
# allocation transfer procedure \
'move_alloc', \
# random number subroutines \
'random_number','random_seed', \
# system environment procedures \
'command_argument_count','cpu_time','date_and_time','get_command','get_command_argument','get_environment_variable','is_iostat_end','is_iostat_eor','system_clock', \
# specific names \
'alog','alog10','amax0','amax1','amin0','amin1','amod','cabs','ccos','cexp','clog','csin','csqrt','dabs','dacos','dasin','datan','dcos','dcosh','ddim','dexp','dint','dlog','dlog10','dmax1','dmin1','dmod','dnint','dsign','dsin','dsinh','dsqrt','dtan','dtanh','float','iabs','idim','idint','idnint','ifix','isign','max0','max1','min0','min1' \
]

Intrinsic_Modules = { \
    'ISO_FORTRAN_ENV': [ 'error_unit', 'file_storage_size', 'input_unit', 'iostat_end', 'iostat_eor', 'numeric_storage_size', 'output_unit', 'character_storage_size' ], \
    'ISO_C_BINDING': [ 'c_associated', 'c_f_pointer', 'c_f_procpointer', 'c_funloc', 'c_loc', 'c_sizeof', 'c_int', 'c_short', 'c_long', 'c_long_long', 'c_signed_char', 'c_size_t', 'c_int8_t', 'c_int16_t', 'c_int32_t', 'c_int64_t', 'c_int_least8_t', 'c_int_least16_t', 'c_int_least32_t', 'c_int_least64_t', 'c_int_fast8_t', 'c_int_fast16_t', 'c_int_fast32_t', 'c_int_fast64_t', 'c_intmax_t', 'c_intptr_t', 'c_ptrdiff_t', 'c_float', 'c_double', 'c_long_double', 'c_float_complex', 'c_double_complex', 'c_long_double_complex', 'c_bool', 'c_char', 'c_null_char', 'c_alert', 'c_backspace', 'c_form_feed', 'c_new_line', 'c_carriage_return', 'c_horizontal_tab', 'c_vertical_tab', 'c_null_ptr', 'c_ptr', 'c_null_funptr', 'c_funptr ' ], \
    'IEEE_EXCEPTIONS': [ 'ieee_flag_type', 'ieee_status_type', 'ieee_support_flag', 'ieee_support_halting', 'ieee_get_flag', 'ieee_get_halting_mode', 'ieee_get_status', 'ieee_set_flag', 'ieee_set_halting_mode', 'ieee_set_status' ], \
    'IEEE_ARITHMETIC': [ 'ieee_class_type', 'ieee_round_type', 'ieee_support_datatype', 'ieee_support_divide', 'ieee_support_denormal', 'ieee_support_inf', 'ieee_support_io', 'ieee_support_nan', 'ieee_support_rounding', 'ieee_support_standard', 'ieee_support_underflow_control', 'ieee_support_sqrt', 'ieee_class', 'ieee_copy_sign', 'ieee_is_finite', 'ieee_is_nan', 'ieee_is_normal', 'ieee_is_negative', 'ieee_is_logb', 'ieee_next_after', 'ieee_rem', 'ieee_rint', 'ieee_scalb', 'ieee_unordered', 'ieee_value', 'ieee_selected_real_kind', 'ieee_get_rounding_mode', 'ieee_get_underflow_mode', 'ieee_set_rounding_mode', 'ieee_set_underflow_mode' ], \
    'IEEE_FEATURES': [ 'ieee_features_type' ] \
}




# kgen_plugin.py

from fortlab.kgutils import ProgramException
import collections

class Kgen_Plugin(object):
    plugin_common = collections.OrderedDict()

    def register(self, msg):
        raise ProgramException('Subclass should implement register function')
